"""ZCC Controller Class."""

from __future__ import annotations

import asyncio
import logging
import socket
import time
import uuid
from pprint import pformat
from typing import Dict, List

from zcc.constants import LEVEL_BY_VERBOSITY, NAME, VERSION
from zcc.description import ControlPointDescription
from zcc.device import ControlPointDevice
from zcc.errors import ControlPointError
from zcc.manufacture_info import ControlPointManufactureInfo
from zcc.protocol import ControlPointProtocol
from zcc.socket import ControlPointSocket
from zcc.watchdog import ControlPointWatchdog


class ControlPoint:
    """Represents the ZCC controller which connects to individual devices"""

    def __init__(
        self,
        description: ControlPointDescription = None,
        timeout: int = 2,
        verbosity: int = 0,
    ):
        self.logger = logging.getLogger("ControlPoint")
        if verbosity > 2:
            verbosity = 2
        self.logger.setLevel(LEVEL_BY_VERBOSITY[verbosity])

        self.host = description.host
        self.port = description.port

        self.logger.info("Setting up %s version %s", NAME, VERSION)

        if not (self.host and self.port):
            raise ControlPointError(
                "Initialisation failed - must provide at least host and port."
            )

        self.brand = description.brand
        self.product = description.product
        self.mac = description.mac
        self.available_tcps = description.available_tcps
        self.api_version = description.api_version
        self.firmware_version = description.firmware_version
        self.num_devices = None
        self.num_control_points = None
        self.network_name = None
        self.uptime = None

        self.timeout = timeout
        self.verbosity = verbosity

        self.devices: Dict[ControlPointDevice] = {}
        self.manufacture_infos: Dict[ControlPointManufactureInfo] = {}

        self.actions_received = 0
        self.manufacture_info_received = 0
        self.properties_received = 0
        self.states_received = 0

        self.device_mac = format(uuid.getnode(), "=012x")
        self.access_token = None
        self.session_authorised = False
        self.session_started = False

        self.socket = None
        self.closed_socket = None

        self.ready = False

        self.loop = asyncio.get_event_loop()
        self.session_authorised: asyncio.Future = None
        self.session_started: asyncio.Future = None
        self.gateway_ready: asyncio.Future = None
        self.manufacture_info_ready: asyncio.Future = None
        self.properties_ready: asyncio.Future = None
        self.actions_ready: asyncio.Future = None
        self.states_ready: asyncio.Future = None

        self.connected = asyncio.Event()
        self.connecting = asyncio.Event()

        self.watchdog_timer: ControlPointWatchdog = None

    @property
    def doors(self) -> List[ControlPointDevice]:
        """Return an array with all doors"""
        return list(
            filter(lambda device: device.type == "garagedoor", self.devices.values())
        )

    @property
    def fans(self) -> List[ControlPointDevice]:
        """Return an array with all fans"""
        return list(filter(lambda device: device.type == "fan", self.devices.values()))

    @property
    def lights(self) -> List[ControlPointDevice]:
        """Return an array with all lights (i.e. switch or dimmer type)"""
        return list(
            filter(
                lambda device: device.type == "light" or device.type == "dimmer",
                self.devices.values(),
            )
        )

    @property
    def outlets(self) -> List[ControlPointDevice]:
        """Return an array with all outlets"""
        return list(
            filter(
                lambda device: device.type == "outlet" or device.type == "switch",
                self.devices.values(),
            )
        )

    @property
    def sensors(self) -> List[ControlPointDevice]:
        """Return an array with all sensors"""
        return list(
            filter(lambda device: device.type == "garagedoor", self.devices.values())
        )

    async def connect(self, fast: bool = False) -> bool:
        """Connect to ZCC, build device table and subscribe to updates"""
        self.logger.info("Connecting to ZCC %s:%d", self.host, self.port)

        self.connecting.set()
        self.connected.clear()

        if self.session_authorised:
            self.session_authorised.cancel()
        self.session_authorised = self.loop.create_future()

        if self.session_started:
            self.session_started.cancel()
        self.session_started = self.loop.create_future()

        if not self.socket:
            self.socket = ControlPointSocket(
                self.host, self.port, timeout=self.timeout, verbosity=self.verbosity
            )

        try:
            await self.socket.connect()
        except ConnectionRefusedError as error:
            self.connecting.clear()
            description = (
                f"Connection refused when connecting to ZCC {self.host}:{self.port}"
            )
            self.disconnect()
            self.logger.error(description)
            raise ControlPointError(description) from error
        except socket.error as error:
            self.connecting.clear()
            description = f"Socket error when connecting to ZCC {self.host}:{self.port}"
            self.logger.error(description)
            self.disconnect()
            raise ControlPointError(description) from error
        except asyncio.CancelledError as error:
            self.connecting.clear()
            description = (
                f"Cancelled error when connecting to ZCC {self.host}:{self.port}"
            )
            self.logger.error(description)
            self.disconnect()
            raise ControlPointError(description) from error
        except Exception as error:
            self.connecting.clear()
            self.logger.error(error)
            self.disconnect()
            raise ControlPointError("Unknown error when connecting to ZCC") from error

        self.socket.subscribe(self)

        await self.socket.sendall(
            ControlPointProtocol.authorise(self.device_mac), response_expected=False
        )

        try:
            self.logger.debug("Waiting for authorisation")
            await asyncio.wait_for(
                self.session_authorised, timeout=ControlPointProtocol.AUTH_TIMEOUT
            )
        except asyncio.exceptions.TimeoutError as error:
            description = f"Unable to authorise session to ZCC {self.host}:{self.port}"
            self.logger.error(description)
            self.disconnect()
            raise ControlPointError(description) from error

        await self.socket.sendall(
            ControlPointProtocol.start(self.device_mac, self.access_token),
            response_expected=False,
        )

        try:
            self.logger.debug("Waiting for session start")
            await asyncio.wait_for(
                self.session_started, timeout=ControlPointProtocol.START_SESSION_TIMEOUT
            )
        except asyncio.exceptions.TimeoutError as error:
            description = f"Unable to start session to ZCC {self.host}:{self.port}"
            self.logger.error(description)
            self.disconnect()
            raise ControlPointError(description) from error

        self.gateway_ready = self.loop.create_future()

        await self.socket.sendall(ControlPointProtocol.inforequest())

        try:
            self.logger.debug("Waiting for gateway properties")
            await asyncio.wait_for(
                self.gateway_ready, timeout=ControlPointProtocol.DEVICE_GET_TIMEOUT
            )
        except asyncio.exceptions.TimeoutError as _:
            description = (
                f"Unable to get gateway properties to ZCC {self.host}:{self.port}"
            )
            self.logger.error(description)

        if not fast:
            await self.__get_control_points()

        await self.socket.sendall(
            ControlPointProtocol.subscribe(), response_expected=False
        )

        await asyncio.sleep(ControlPointProtocol.SUBSCRIBE_TIMEOUT)

        self.ready = True

        self.connected.set()
        self.connecting.clear()

        self.logger.info(
            "Connected to ZCC %s:%d with %d/%d/%d actions/properties/states",
            self.host,
            self.port,
            self.actions_received,
            self.states_received,
            self.properties_received,
        )

    def __del__(self):
        self.disconnect()

    def describe(self) -> str:
        """Return a string representation of ZCC including devices"""
        header = "+" + "-" * 130 + "+"
        if self.host:
            description = header + "\n"
            description += (
                "| Device mac:   %35s          Brand:       %8s              Product:            %8s       |\n"
                % (
                    self.mac if self.mac else "n/a",
                    self.brand if self.brand else "n/a",
                    self.product if self.product else "n/a",
                )
            )
            description += (
                "| Host:         %35s          Port:        %8d              API:                %8s       |\n"
                % (
                    self.host if self.host else "n/a",
                    self.port if self.port else "n/a",
                    self.api_version if self.api_version else "n/a",
                )
            )
            description += (
                "| Firmware:     %35s          Num Devices: %8s              Num Control Points: %8s       |\n"
                % (
                    self.firmware_version if self.firmware_version else "n/a",
                    self.num_devices if self.num_devices else "n/a",
                    self.num_control_points if self.num_control_points else "n/a",
                )
            )
            description += (
                "| Network Name: %35s          Uptime:     %8ss              Tcps:               %8s       |\n"
                % (
                    self.network_name if self.network_name else "n/a",
                    self.uptime if self.uptime else "n/a",
                    str(self.available_tcps) if self.available_tcps else "n/a",
                )
            )
            description += header + "\n"
            description_details = []
            for key in self.manufacture_infos:
                description_details.append(self.manufacture_infos[key].describe())
            for key in self.devices:
                description_details.append(self.devices[key].describe())
            description_details.sort()
            description += "".join(description_details)

            return description
        else:
            return "ControlPoint: not found"

    def disconnect(self):
        """Disconnect from zimi controller"""
        self.ready = False
        if self.socket:
            self.socket.close()
            self.socket = None

    async def __get_control_points(self):
        """Get initial control point data from controller."""

        self.manufacture_info_ready = self.loop.create_future()
        self.manufacture_info_received = 0
        self.properties_ready = self.loop.create_future()
        self.properties_received = 0
        self.actions_ready = self.loop.create_future()
        self.actions_received = 0
        self.states_ready = self.loop.create_future()
        self.states_received = 0

        self.logger.debug("Getting manufacture_info")
        await self.socket.sendall(
            ControlPointProtocol.get("manufacture_info"), response_expected=False
        )

        try:
            await asyncio.wait_for(
                self.manufacture_info_ready,
                timeout=ControlPointProtocol.DEVICE_GET_TIMEOUT,
            )
        except asyncio.exceptions.TimeoutError as error:
            raise ControlPointError(
                "ZCC connection failed - didn't receive manufacture_info."
            ) from error

        await asyncio.sleep(ControlPointProtocol.STEP_TIMEOUT)

        self.logger.debug("Getting initial device properties")
        await self.socket.sendall(
            ControlPointProtocol.get("properties"), response_expected=False
        )

        try:
            await asyncio.wait_for(
                self.properties_ready, timeout=ControlPointProtocol.DEVICE_GET_TIMEOUT
            )
        except asyncio.exceptions.TimeoutError as error:
            raise ControlPointError(
                "ZCC connection failed - didn't receive any properties."
            ) from error

        await asyncio.sleep(ControlPointProtocol.STEP_TIMEOUT)

        self.logger.debug("Getting initial device actions")
        await self.socket.sendall(
            ControlPointProtocol.get("actions"), response_expected=False
        )

        try:
            await asyncio.wait_for(
                self.actions_ready, timeout=ControlPointProtocol.DEVICE_GET_TIMEOUT
            )
        except asyncio.exceptions.TimeoutError as error:
            raise ControlPointError(
                "ZCC connection failed - didn't receive any actions."
            ) from error

        await asyncio.sleep(ControlPointProtocol.STEP_TIMEOUT)

        self.logger.debug("Getting initial device states")
        await self.socket.sendall(
            ControlPointProtocol.get("states"), response_expected=False
        )

        try:
            await asyncio.wait_for(
                self.states_ready, timeout=ControlPointProtocol.DEVICE_GET_TIMEOUT
            )
        except asyncio.exceptions.TimeoutError as error:
            raise ControlPointError(
                "ZCC connection failed - didn't receive any states."
            ) from error

        await asyncio.sleep(ControlPointProtocol.STEP_TIMEOUT)

        for key in self.devices.keys():
            identifier_msb = self.devices[key].identifier.split("_")[0]
            if self.manufacture_infos[identifier_msb]:
                self.devices[key].manufacture_info = self.manufacture_infos[
                    identifier_msb
                ]

    async def __get_states(self):
        """Get latest state data from controller and reset watchdog."""

        self.states_ready = self.loop.create_future()

        self.logger.debug("Refreshing device states")
        await self.socket.sendall(ControlPointProtocol.get("states"))

        try:
            await asyncio.wait_for(
                self.states_ready, timeout=ControlPointProtocol.DEVICE_GET_TIMEOUT
            )
            if self.watchdog_timer:
                self.watchdog_timer.reset()
        except asyncio.exceptions.TimeoutError as error:
            self.logger.error(
                "Unable to refresh connection to ZCC - will re-connect: %s\n", error
            )
            await self.re_connect()

    def print_description(self):
        """Print description of the ZCC controller."""
        print(self.describe())

    async def notify(self, notifier):
        """Receive a notification of an updated object.
        Pull the data off the queue.   If None is received then
        assume the socket has been closed and needs re-opening."""

        response = notifier.get()

        if not response:
            if self.ready:
                await self.re_connect()
            return

        self.logger.debug("notify() received:\n%s", response)

        if response.get(ControlPointProtocol.AUTH_APP_FAILED, None):
            self.logger.error("Authorisation failed\n%s", pformat(response))
        if response.get(ControlPointProtocol.AUTH_APP_SUCCESS, None):
            self.logger.debug("Authorisation success\n%s", pformat(response))
            self.access_token = response[ControlPointProtocol.AUTH_APP_SUCCESS][
                "accessToken"
            ]
            self.session_authorised.set_result(True)
        if response.get(ControlPointProtocol.START_SESSION_FAILED, None):
            self.logger.error("Start session failed\n%s", pformat(response))
        if response.get(ControlPointProtocol.START_SESSION_SUCCESS, None):
            self.logger.debug("Start session success\n%s", pformat(response))
            self.session_started.set_result(True)
        if response.get(ControlPointProtocol.CONTROLPOINT_ACTIONS, None):
            self.__update_control_points(
                response.get(ControlPointProtocol.CONTROLPOINT_ACTIONS, None), "actions"
            )
        if response.get(ControlPointProtocol.CONTROLPOINT_PROPERTIES, None):
            self.__update_control_points(
                response.get(ControlPointProtocol.CONTROLPOINT_PROPERTIES, None),
                "properties",
            )
        if response.get(ControlPointProtocol.CONTROLPOINT_MANUFACTURE_INFO, None):
            self.__update_manufacture_info(
                response.get(ControlPointProtocol.CONTROLPOINT_MANUFACTURE_INFO, None)
            )
        if response.get(ControlPointProtocol.CONTROLPOINT_STATES, None):
            self.__update_control_points(
                response.get(ControlPointProtocol.CONTROLPOINT_STATES, None), "states"
            )
        if response.get(ControlPointProtocol.CONTROLPOINT_STATES_EVENTS, None):
            self.__update_control_points(
                response.get(ControlPointProtocol.CONTROLPOINT_STATES_EVENTS, None),
                "states",
            )
        if response.get(ControlPointProtocol.GATEWAY_PROPERTIES, None):
            self.gateway_ready.set_result(True)
            self.__update_properties(
                response.get(ControlPointProtocol.GATEWAY_PROPERTIES, None)
            )
        if response.get(ControlPointProtocol.ZCC_STATUS, None):
            self.logger.error("ZCC status message\n%s", pformat(response))
            await self.re_connect(reset=True)

    async def re_connect(self, reset: bool = False):
        """Re-connect to a new socket and resend any queued messages."""

        if reset:
            self.logger.error(
                "Preparing re-connect to new socket after ZCC intiated reset"
            )
        else:
            self.logger.error(
                "Preparing re-connect to new socket as existing socket closed"
            )

        if self.watchdog_timer:
            self.logger.debug("Pausing existing watchdog timer during re-connection")
            self.watchdog_timer.pause()

        self.socket.unsubscribe(self)
        self.socket.close()

        self.closed_socket = self.socket

        self.ready = False

        while not self.ready:
            self.socket = None

            try:
                self.logger.debug("Re-connecting to ZCC with new socket")
                await self.connect(fast=True)

                while True:
                    message = self.closed_socket.unsent()
                    if message:
                        self.logger.debug("Re-sending message:\n%s", message)
                        await self.socket.sendall(message)
                    else:
                        break
                self.closed_socket = None
            except ControlPointError as error:
                self.logger.error(
                    "Unable to re-connect to ZCC failed with ControlPointError: %s - will retry in %d",
                    error,
                    ControlPointProtocol.RETRY_TIMEOUT,
                )
                await asyncio.sleep(ControlPointProtocol.RETRY_TIMEOUT)

        if self.watchdog_timer:
            self.logger.debug("Re-starting existing watchdog timer after re-connection")
            self.watchdog_timer.reset()

    async def set(self, identifier: str, action: str, params: object = None):
        """Sends an action for a device."""

        while self.connecting.is_set():
            self.logger.error(
                "Controller not ready to accept commands - in process of re-connect"
            )
            await asyncio.sleep(1)

        while not self.connected.is_set():
            self.logger.error(
                "Controller not ready to accept commands - attempting to re-connect"
            )
            try:
                await self.re_connect()
            except ControlPointError as error:
                self.logger.error(
                    "Unable to re-connect to ZCC failed with ControlPointError: %s - will retry",
                    error,
                )

        message = ControlPointProtocol.set(identifier, action, params)
        success = False
        while not success:
            success = await self.socket.sendall(message)
            self.logger.debug(
                "Sending %s request to %s (%s)",
                action,
                self.devices[identifier].location,
                identifier,
            )

    def __str__(self):
        return pformat(vars(self)) + "\n"

    def start_watchdog(self, timer: int = 1800):
        """Start a periodic timeout that resets every time a status update is received."""

        if self.watchdog_timer:
            self.stop_watchdog()

        self.logger.debug("Starting Watchdog for %s seconds", timer)
        self.watchdog_timer = ControlPointWatchdog(timer, self.trigger_watchdog)
        self.watchdog_timer.start()

    def stop_watchdog(self):
        """Stop the periodic timeout."""

        self.logger.debug("Stopping existing Watchdog")

        if self.watchdog_timer:
            self.watchdog_timer.cancel()
            self.watchdog_timer = None

    async def trigger_watchdog(self):
        """Trigger the watchdog function - which will reset the connection."""

        self.logger.debug(
            "Triggering the watchdog timer - will fetch new states to refresh connection"
        )
        await self.__get_states()

    def __update_control_points(self, devices, target):
        """Update control point target with JSON data for all devices."""

        for device in devices:
            updates_made = False
            identifier = device["id"]
            if not self.devices.get(identifier):
                self.devices[identifier] = ControlPointDevice(self, identifier)
            if (
                "actions" in target
                and self.devices[identifier].actions != device[target]
            ):
                self.devices[identifier].actions = device[target]
                self.actions_received += 1
                updates_made = True
            if (
                "properties" in target
                and self.devices[identifier].properties != device[target]
            ):
                self.devices[identifier].properties = device[target]
                self.properties_received += 1
                updates_made = True
            if "states" in target and self.devices[identifier].states != device[target]:
                self.devices[identifier].states = device[target]
                self.states_received += 1
                updates_made = True
            if updates_made:
                if self.watchdog_timer:
                    self.watchdog_timer.reset()
                self.logger.debug(
                    "Received %d/%d actions; %d/%d properties; %d/%d states",
                    self.actions_received,
                    self.num_control_points,
                    self.properties_received,
                    self.num_control_points,
                    self.states_received,
                    self.num_control_points,
                )
                self.logger.debug(
                    "Received %s update for %s (%s):\n%s",
                    target,
                    self.devices[identifier].location,
                    identifier,
                    device[target],
                )
                self.devices[identifier].notify_observers()

        if (
            self.actions_received >= self.num_control_points
            and not self.actions_ready.done()
        ):
            self.actions_ready.set_result(True)
        if (
            self.properties_received >= self.num_control_points
            and not self.properties_ready.done()
        ):
            self.properties_ready.set_result(True)
        if (
            self.states_received >= self.num_control_points
            and not self.states_ready.done()
        ):
            self.states_ready.set_result(True)

    def __update_manufacture_info(self, infos):
        """Update manufacture_info target with JSON data for all devices."""

        for info in infos:
            identifier = info.get("id", None)
            elements = info.get("info", None)
            if identifier and elements:
                self.manufacture_info_received += 1
                self.manufacture_infos[identifier] = ControlPointManufactureInfo(
                    identifier=identifier,
                    manufacturer=elements.get("manufacturer", None),
                    model=elements.get("model", None),
                    hwVersion=elements.get("hwVersion", None),
                    firmwareVersion=elements.get("firmwareVersion", None),
                )

                self.logger.debug(
                    "Received %d/%d: %s",
                    self.manufacture_info_received,
                    self.num_devices,
                    self.manufacture_infos[identifier],
                )

        if self.watchdog_timer:
            self.watchdog_timer.reset()
        if (
            self.manufacture_info_received >= self.num_devices
            and not self.manufacture_info_ready.done()
        ):
            self.manufacture_info_ready.set_result(True)

    def __update_properties(self, properties):
        """Updates zcc gateway properties"""

        properties = properties.get("properties")

        if properties:
            self.brand = properties["brand"]
            self.product = properties["product"]
            self.mac = properties["mac"]
            self.available_tcps = properties["availableTcps"]
            if api_version := properties.get("apiVersion"):
                self.api_version = api_version
            if firmware_version := properties.get("firmwareVersion"):
                self.firmware_version = firmware_version
            self.num_devices = properties["numberOfDevices"]
            self.num_control_points = properties["numberOfControlPoints"]
            self.network_name = properties["networkName"]
            self.uptime = int(time.time()) - properties["uptime"]



import socket
import pytest
from unittest.mock import Mock
from mockito import when, unstub

from zcc.socket import ControlPointSocket

TEST_HOST = '10.0.0.1'
TEST_PORT = 4567
TEST_TIMEOUT = 5


@pytest.fixture
def mock_tcp_socket():
    mock_socket = Mock(spec=socket.socket)
    return mock_socket


def test_socket_create(mock_tcp_socket, when):

    when(socket).socket(socket.AF_INET,
                        socket.SOCK_STREAM)\
        .thenReturn(mock_tcp_socket)

    controller_socket = ControlPointSocket(TEST_HOST, TEST_PORT)

    assert controller_socket.host == TEST_HOST
    assert controller_socket.port == TEST_PORT
    assert controller_socket.sock == mock_tcp_socket

    controller_socket.close()
    assert controller_socket.sock == None


def test_socket_create_timeout(mock_tcp_socket, when):

    when(socket).socket(socket.AF_INET,
                        socket.SOCK_STREAM)\
        .thenReturn(mock_tcp_socket)

    controller_socket = ControlPointSocket(
        TEST_HOST, TEST_PORT, timeout=TEST_TIMEOUT)

    assert controller_socket.host == TEST_HOST
    assert controller_socket.port == TEST_PORT
    assert controller_socket.sock == mock_tcp_socket

    assert mock_tcp_socket.settimeout.call_count == 1
    assert mock_tcp_socket.settimeout.call_args.args == (TEST_TIMEOUT,)

    controller_socket.close()
    assert controller_socket.sock == None



'''Setup Test Server.'''
import pytest

from zcc.controller import ControlPoint, ControlPointError
from tests.server import TestServer


@pytest.fixture
def test_server():
    '''Create (and delete) a Test Server'''
    return TestServer()



'''ControlPoint error class'''


class ControlPointError(Exception):
    '''Represents a ZCC controller error.'''

class ControlPointCannotConnectError(ControlPointError):
    '''Represents a connect connect error when connecting to zcc.'''
    
class ControlPointInvalidHostError(ControlPointError):
    '''Represents an invalid host error when connecting to zcc.'''
class ControlPointConnectionRefusedError(ControlPointError):
    '''Represents a connection refused when connecting to zcc.'''

class ControlPointTimeoutError(ControlPointError):
    '''Represents a connection timeout when connecting to zcc.'''


'''ZCC Watchdog Timer Class'''

import asyncio
import logging


class ControlPointWatchdog:
    '''A class to create a timer with callback.'''

    def __init__(self, timeout: int, callback):

        self.logger = logging.getLogger('ControlPointWatchdog')

        self._busy = False
        self._timeout = timeout
        self._callback = callback
        self._task = None
        self.logger.debug(
            "Created watchdog timer for %d seconds", self._timeout)

    async def _job(self):
        await asyncio.sleep(self._timeout)
        self._busy = True
        await self._callback()
        self._busy = False

    def cancel(self):
        '''Cancel the timer.'''
        if not self._busy:
            self._task.cancel()

    def pause(self):
        '''Pause the timer'''
        if self._task:
            self.cancel()

    def reset(self):
        '''Reset the timer.'''
        if self._task:
            self.cancel()
        self.start()

    def start(self):
        '''Start the timer.'''
        self._task = asyncio.ensure_future(self._job())



"""ControlPointDevice Class."""

from __future__ import annotations

from pprint import pformat

from zcc.errors import ControlPointError
from zcc.manufacture_info import ControlPointManufactureInfo

OPEN_AND_CLOSE_PERCENTAGE_TOLERANCE = 10


class ControlPointDevice:
    """Represents any ControlPointDevice."""

    def __init__(self, controller, identifier: str):
        """Create a new device associated with a controller and identifed by identifier"""
        self.controller = controller
        self.identifier = identifier
        self.manufacture_info: ControlPointManufactureInfo
        self.actions = {}
        self.properties = {}
        self.states = {}

        self._target_percentage = None
        self._observers = []

    def __str__(self):
        return pformat(vars(self))

    async def __action(self, action: str, params: object = None) -> bool:
        try:
            if self.actions["actions"][action]:
                await self.controller.set(self.identifier, action, params=params)
        except Exception as exception_info:
            raise ControlPointError(
                action + " is not supported for " + self.identifier
            ) from exception_info

    @property
    def __actions(self) -> str:
        """Gets a descriptive string of what actions are supported"""
        description = "{ "
        try:
            if self.actions["actions"].get("TurnOn", None):
                description += "TurnOn TurnOff "
            if self.actions["actions"].get("SetBrightness", None):
                description += "SetBrightness "
            if self.actions["actions"].get("OpenDoor", None):
                description += "OpenDoor CloseDoor "
            if self.actions["actions"].get("OpenToPercentage", None):
                description += "OpenToPercentage "
            if self.actions["actions"].get("SetFanSpeed", None):
                description += "SetFanSpeed "
        except KeyError:
            pass
        description += "}"
        if description == "{ }":
            description = ""
        return description

    @property
    def __states(self) -> str:
        """Gets a descriptive string of the device state"""
        description = ""
        try:
            key = list(self.states["controlState"].keys())[0]
            state = self.states["controlState"][key]
            description += "On" if state.get("isOn", False) else "Off"
            brightness = state.get("brightness", None)
            if brightness:
                description += "/" + str(brightness)
            fan_speed = state.get("fanspeed", None)
            if fan_speed:
                description += "/" + str(fan_speed)
        except IndexError:
            pass
        except KeyError:
            pass
        return description

    @property
    def battery_level(self) -> int | None:
        """Return the battery level of an attached sensor."""
        try:
            state = self.states["controlState"]["sensor"]
            return state.get("batterylevel", 0)
        except KeyError:
            return None

    @property
    def brightness(self) -> int | None:
        """Returns brightness from 0 to 100 or None."""
        try:
            key = list(self.states["controlState"].keys())[0]
            state = self.states["controlState"][key]
            brightness = state.get("brightness", None)
            if brightness:
                return brightness
        except KeyError:
            return False

    async def close_door(self):
        """CloseDoor if the action is supported"""
        self._target_percentage = 0
        await self.__action("CloseDoor")

    def describe(self) -> str:
        """Returns a description of a device"""
        return "%-40s   %-38.38s %-8.8s %-8s %s\n" % (
            self.identifier,
            self.location,
            self.type,
            self.__states,
            self.__actions,
        )

    @property
    def door_temp(self) -> int | None:
        """Return the external temperature of an attached sensor."""
        try:
            state = self.states["controlState"]["sensor"]
            return state.get("doortemp", 0)
        except KeyError:
            return None

    async def fade(self, brightness, timeperiod):
        """SetBrightness if the action is supported"""
        await self.__action(
            "SetBrightness",
            params={"brightness": int(brightness), "timeperiod": int(timeperiod)},
        )

    @property
    def fanspeed(self) -> int | None:
        """Returns fanspeed from 0 to 7 or None."""
        try:
            key = list(self.states["controlState"].keys())[0]
            state = self.states["controlState"][key]
            fanspeed = state.get("fanspeed", None)
            if fanspeed:
                return fanspeed
        except KeyError:
            return False

    @property
    def garage_temp(self) -> int | None:
        """Return the internal garage temperature of an attached sensor."""
        try:
            state = self.states["controlState"]["sensor"]
            return state.get("garagetemp", 0)
        except KeyError:
            return None

    @property
    def garage_humidity(self) -> int | None:
        """Return the internal garage humidity of an attached sensor."""
        try:
            state = self.states["controlState"]["sensor"]
            return state.get("garagehumidity", 0)
        except KeyError:
            return None

    @property
    def is_closing(self) -> bool:
        """Returns True if door is closing."""
        if self.percentage and self._target_percentage:
            return self.percentage > self._target_percentage and not self.is_closed
        return False

    @property
    def is_closed(self) -> bool:
        """Returns True if door is closed."""

        if self.percentage:
            return self.percentage <= OPEN_AND_CLOSE_PERCENTAGE_TOLERANCE
        return True

    @property
    def is_connected(self) -> bool:
        """Returns True if connected is on.
        When a device has been disconnected from the mesh it show False."""

        try:
            return self.states["isConnected"]
        except KeyError:
            return False

    @property
    def is_off(self) -> bool:
        """Returns True if status is off."""

        return not self.is_on

    @property
    def is_on(self) -> bool:
        """Returns True if status is on."""

        try:
            key = list(self.states["controlState"].keys())[0]
            state = self.states["controlState"][key]
            return state.get("isOn", False)
        except KeyError:
            return False

    @property
    def is_opening(self) -> bool:
        """Returns True if door is opening and is NOT already open."""

        if self.percentage and self._target_percentage:
            return self.percentage < self._target_percentage and not self.is_open
        return False

    @property
    def is_open(self) -> bool:
        """Returns True if door is open."""

        if self.percentage:
            return self.percentage >= 100 - OPEN_AND_CLOSE_PERCENTAGE_TOLERANCE
        return False

    @property
    def location(self) -> str:
        """Gets a descriptive string of the device location"""
        description = self.properties.get("name", "-") + "/"
        description += self.properties.get("roomName", "-")
        return description

    @property
    def device_name(self) -> str:
        """Gets a device name."""
        return self.manufacture_info.name

    @property
    def name(self) -> str:
        """Gets a descriptive string of the device name"""
        return self.properties.get("name", "-")

    def notify_observers(self):
        """Notify all observers that a state change has occurred."""
        for obs in self._observers:
            obs.notify(self)

    async def open_door(self):
        """OpenDoor if the action is supported"""
        self._target_percentage = 100
        await self.__action("OpenDoor")

    async def open_to_percentage(self, percentage):
        """OpenToPercentage if the action is supported"""
        self._target_percentage = percentage
        await self.__action(
            "OpenToPercentage", params={"openpercentage": int(percentage)}
        )

    @property
    def percentage(self) -> int | None:
        """Return the open to percentage"""
        try:
            key = list(self.states["controlState"].keys())[0]
            state = self.states["controlState"][key]
            return state.get("openpercentage", 0)
        except KeyError:
            return None

    def print_description(self):
        """Print device description."""
        print(self.describe())

    @property
    def room(self) -> str:
        """Gets a descriptive string of the device room"""
        return self.properties.get("roomName", "-")

    async def set_brightness(self, brightness):
        """SetBrightness if the action is supported"""
        await self.__action("SetBrightness", params={"brightness": int(brightness)})

    async def set_fanspeed(self, fanspeed):
        """SetFanSpeed if the action is supported"""
        await self.__action("SetFanSpeed", params={"fanspeed": int(fanspeed)})

    def subscribe(self, observer):
        """Subscribe an observer object for state changes.
        Observer object must include notify(self, observable, *args, **kwargs) method."""
        self._observers.append(observer)

    async def turn_on(self):
        """TurnOn the device if the action is supported"""
        await self.__action("TurnOn")

    async def turn_off(self):
        """TurnOff the device if the action is supported"""
        await self.__action("TurnOff")

    @property
    def type(self) -> str:
        """Gets a descriptive string of the device type"""
        return self.properties.get("controlPointType", "-")

    def unsubscribe(self, observer):
        """Unsubscribe an observer object."""
        self._observers.remove(observer)



"""ZCC Discovery Service Class"""

from __future__ import annotations

import asyncio
from enum import StrEnum
import json
from json.decoder import JSONDecodeError
import logging
import socket
from typing import Tuple

from zcc.constants import LEVEL_BY_VERBOSITY
from zcc.controller import ControlPoint
from zcc.description import ControlPointDescription
from zcc.errors import ( ControlPointError,
                     ControlPointConnectionRefusedError,
                     ControlPointCannotConnectError,
                     ControlPointInvalidHostError,
                     ControlPointTimeoutError)
from zcc.protocol import ControlPointProtocol

class ControlPointDiscoveryProtocol(asyncio.DatagramProtocol):
    """Listens for ZCC announcements on the defined UDP port."""

    def __init__(
        self, discovery_complete: asyncio.Future | None, discovery_result: object
    ) -> None:
        super().__init__()
        self.discovery_complete: asyncio.Future | None = discovery_complete
        self.discovery_result = discovery_result
        self.transport: asyncio.transports.DatagramTransport = None
        self.logger = logging.getLogger("ControlPointDiscoveryService")

    def connection_lost(self, exc) -> None:
        self.transport.close()
        return super().connection_lost(exc)

    def connection_made(self, transport: asyncio.transports.DatagramTransport) -> None:
        self.transport = transport
        return super().connection_made(transport)

    def datagram_received(self, data: bytes, addr: Tuple[str, int]) -> None:
        data = str(data.decode("UTF-8"))
        self.logger.debug("datagram_received() from %s\n%s", str(addr), data)

        lines = data.split("\n")
        for line in lines:
            try:
                if response := json.loads(line):
                    result = ControlPointDescription()
                    result.brand = response["brand"]
                    result.product = response["product"]
                    result.mac = response["mac"]
                    result.host = addr[0]
                    result.port = response["tcp"]
                    result.available_tcps = response["availableTcps"]
                    if api_version := response.get("apiVersion"):
                        result.api_version = api_version
                    if firmware_version := response.get("firmwareVersion"):
                        result.firmware_version = firmware_version
                    self.discovery_result.append(result)
                    if self.discovery_complete:
                        self.discovery_complete.set_result(True)
            except JSONDecodeError:
                break

        return super().datagram_received(data, addr)


class ControlPointDiscoveryService:
    """Provides a ZCC discovery service to discover ZIMI controllers on the local LAN."""

    def __init__(self, verbosity: int = 0):
        self.logger = logging.getLogger("ControlPointDiscoveryService")
        if verbosity > 2:
            verbosity = 2
        self.logger.setLevel(LEVEL_BY_VERBOSITY[verbosity])

        self.loop = asyncio.get_event_loop()
        self.discovery_complete = self.loop.create_future()
        self.discovery_result: list[ControlPointDescription] = []
        self.never_completes = self.loop.create_future()
        self.validation_result: ControlPointDescription

    async def discover(
        self, wait_for_all: bool = False
    ) -> ControlPointDescription | list[ControlPointDescription]:
        """Discover local ZIMI controllers on LAN and return (host,port)."""

        if wait_for_all:
            self.discovery_complete = None

        transport, _ = await self.loop.create_datagram_endpoint(
            lambda: ControlPointDiscoveryProtocol(
                self.discovery_complete, self.discovery_result
            ),
            local_addr=("0.0.0.0", ControlPointProtocol.UDP_RECV_PORT),
        )

        send_socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        send_socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        send_socket.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)

        server_address = ("255.255.255.255", ControlPointProtocol.UDP_SEND_PORT)
        message = ControlPointProtocol.discover()

        send_socket.sendto(message.encode(), server_address)
        self.logger.info("Sending discovery message on local network")

        if wait_for_all:
            try:
                await asyncio.wait_for(self.never_completes, timeout=10)
            except asyncio.exceptions.TimeoutError as error:
                transport.close()
                if len(self.discovery_result) <= 0:
                    raise ControlPointError(
                        "Failure - Unable to discover ZCC by UDP broadcast."
                    ) from error
                self.logger.info(
                    "Success - discovered ZIMI controllers:",
                )
                for result in self.discovery_result:
                    self.logger.info(result)
                return self.discovery_result
        else:
            try:
                await asyncio.wait_for(self.discovery_complete, timeout=10)
                transport.close()
                self.logger.info("Success - discovered ZIMI controller:")
                self.logger.info(
                    self.discovery_result[0],
                )
                return self.discovery_result[0]
            except asyncio.exceptions.TimeoutError as error:
                transport.close()
                raise ControlPointError(
                    "Failure - Unable to discover ZCC by UDP broadcast."
                ) from error

    async def discovers(self) -> list[ControlPointDescription]:
        """Discover all local zimi controllers."""
        return await self.discover(wait_for_all=True)

    async def validate_connection(
        self, host: str, port: int
    ) -> ControlPointDescription:
        """Validate ability to connect and close a connection.

        Returns ControlPointDescription if OK.
        """

        try:
            socket.gethostbyname(host)
            s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            s.settimeout(10)
            try:
                s.connect((host, port))
                s.close()
            except ConnectionRefusedError as e:
                raise ControlPointConnectionRefusedError() from e
            except TimeoutError as e:
                raise ControlPointTimeoutError() from e
            except socket.gaierror as e:
                raise ControlPointCannotConnectError() from e
        except socket.gaierror as e:
            raise ControlPointInvalidHostError() from e

        api = ControlPoint(ControlPointDescription(host=host, port=port))

        try:
            await api.connect(fast=True)
        except ControlPointError as e:
            raise ControlPointCannotConnectError() from e

        self.validation_result = ControlPointDescription(
            brand=api.brand,
            product=api.product,
            host=host,
            port=port,
            mac=api.mac,
            available_tcps=api.available_tcps,
            api_version=api.api_version,
            firmware_version=api.firmware_version
        )

        api.disconnect()
        
        self.logger.info(self.validation_result)

        return self.validation_result



"""ControlPointManufactureInfo Class."""

from dataclasses import dataclass


@dataclass
class ControlPointManufactureInfo:
    """Data class to store ControlPoint manufacture_info."""

    identifier: str = None
    manufacturer: str = None
    model: str = None
    hwVersion: str = None
    firmwareVersion: str = None

    def describe(self) -> str:
        """Returns a description of a device"""
        return "%-40s %-40.40s %-8.8s          %s\n" % (
            self.identifier,
            f"{self.model} ({self.name})",
            "device",
            f"( hw={self.hwVersion}, fw={self.firmwareVersion} )",
        )

    @property
    def name(self) -> str:
        """Returns simplified name of the device."""

        if not self.model:
            return "N/A"

        if "Fan" in self.model:
            return "Fan"
        if "GPO" in self.model:
            return "Outlet"
        if "Switch" in self.model:
            return "Switch"

        return self.model



'''ZCC Support Library to support ZIMI based home network elements'''
import logging
import sys

from .controller import ControlPoint
from .description import ControlPointDescription
from .discovery import ControlPointDiscoveryService
from .errors import ( ControlPointError,
                     ControlPointConnectionRefusedError,
                     ControlPointCannotConnectError,
                     ControlPointInvalidHostError,
                     ControlPointTimeoutError)

logging.basicConfig(stream=sys.stderr)



"""Static methods for the Zimi Controller protocol."""

import json

from .constants import APP_ID, APP_TOKEN


class ControlPointProtocol:
    """Static methods for the Zimi Controller protocol."""

    AUTH_APP_FAILED = "auth_app_failed"
    AUTH_APP_SUCCESS = "auth_app_success"
    AUTH_TIMEOUT = 30
    START_SESSION_FAILED = "start_session_failed"
    START_SESSION_SUCCESS = "start_session_success"
    START_SESSION_TIMEOUT = 30
    DEVICE_GET_TIMEOUT = 10
    CONTROLPOINT_PROPERTIES = "controlpoint_properties"
    CONTROLPOINT_STATES = "controlpoint_states"
    CONTROLPOINT_ACTIONS = "controlpoint_actions"
    CONTROLPOINT_MANUFACTURE_INFO = "controlpoint_manufacture_info"
    CONTROLPOINT_SETACTIONS = "controlpoint_setactions"
    CONTROLPOINT_STATES_EVENTS = "controlpoint_states_events"
    ZCC_STATUS = "zcc_status"

    SUBSCRIBE_TIMEOUT = 2
    GATEWAY_PROPERTIES = "gateway_properties"
    GATEWAY_PROPERTIES_TIMEOUT = 10

    STEP_TIMEOUT = 0

    RETRY_TIMEOUT = 90

    RESET_TIMEOUT = 90

    UDP_SEND_PORT = 5001
    UDP_RECV_PORT = 5002

    @classmethod
    def authorise(cls, device_mac: str) -> str:
        """Returns the authorise message."""

        request = {
            "request": {
                "type": "auth_app",
                "params": {
                    "appId": APP_ID,
                    "appToken": APP_TOKEN,
                    "deviceMac": device_mac,
                },
            }
        }

        return json.dumps(request)

    @classmethod
    def discover(cls) -> str:
        """Return the discover message to be broadcast on LAN."""
        return "ZIMI"

    @classmethod
    def get(cls, target: str) -> str:
        """Returns the get messages assocated with a target"""
        request = {"request": {"path": "api/v1/controlpoint/", "method": "GET"}}
        if target in ("actions", "manufacture_info", "properties", "states"):
            request["request"]["path"] += target
            return json.dumps(request)
        else:
            raise KeyError(
                'Only "actions", "manufacture_info", "properties" and "states" are supported'
            )

    @classmethod
    def set(cls, identifier: str, action: str, params: object) -> str:
        """Returns the set messages assocated with an action"""
        request = {"request": {"path": "api/v1/controlpoint/actions", "method": "POST"}}
        if action in ("TurnOn", "TurnOff", "OpenDoor", "CloseDoor"):
            body = {"actions": [{"id": identifier, "action": action}]}
            request["request"]["body"] = body
        elif (
            action in ("Fade", "OpenToPercentage", "SetBrightness", "SetFanSpeed")
            and params
        ):
            body = {
                "actions": [
                    {"id": identifier, "action": action, "actionParams": params}
                ]
            }
            request["request"]["body"] = body
        else:
            raise KeyError("Combination of input parameters are not supported")
        return json.dumps(request)

    @classmethod
    def start(cls, device_mac: str, access_token: str) -> str:
        """Returns the start session message."""

        request = {
            "request": {
                "type": "start_session",
                "params": {
                    "appId": APP_ID,
                    "deviceMac": device_mac,
                    "accessToken": access_token,
                },
            }
        }

        return json.dumps(request)

    @classmethod
    def subscribe(cls) -> str:
        """Returns the message assocated with state subscribe"""
        request = {
            "request": {
                "path": "api/v1/subscribe/controlpoint/states",
                "method": "POST",
            }
        }
        return json.dumps(request)

    @classmethod
    def unsubscribe(cls) -> str:
        """Returns the message assocated with state subscribe"""
        request = {
            "request": {
                "path": "api/v1/unsubscribe/controlpoint/states",
                "method": "POST",
            }
        }
        return json.dumps(request)

    @classmethod
    def inforequest(cls) -> str:
        """Returns the message assocated with gateway properties request"""
        request = {"request": {"path": "api/v1/gateway/properties", "method": "GET"}}
        return json.dumps(request)



import unittest

from zcc.device import ControlPointDevice


class DeviceTest(unittest.TestCase):

    identifier = 'test_device_id'

    def test_init(self):
        device = ControlPointDevice(None, DeviceTest.identifier)
        self.assertEqual(device.controller, None)
        self.assertEqual(device.identifier, DeviceTest.identifier)
        self.assertEqual(device.actions, {})
        self.assertEqual(device.properties, {})
        self.assertEqual(device.states, {})

    def test_add_action(self):
        device = ControlPointDevice(None, DeviceTest.identifier)
        device.actions = {'TurnOn': {'actionParams': {}},
                          'TurnOff': {'actionParams': {}}}
        print(device)
        self.assertEqual(device.controller, None)
        self.assertEqual(device.identifier, DeviceTest.identifier)

    def test_add_properties(self):
        device = ControlPointDevice(None, DeviceTest.identifier)
        device.properties = {'name': 'Entry Pendant',
                             'controlPointType': 'switch', 'roomId': 2, 'roomName': 'Front Door '}
        self.assertEqual(device.controller, None)
        self.assertEqual(device.identifier, DeviceTest.identifier)

    def test_add_states(self):
        device = ControlPointDevice(None, DeviceTest.identifier)
        device.states = {'controlState': {
            'switch': {...}}, 'isConnected': True}
        self.assertEqual(device.controller, None)
        self.assertEqual(device.identifier, DeviceTest.identifier)


if __name__ == '__main__':
    unittest.main()



"""A setuptools based setup module."""

import pathlib
from setuptools import setup, find_packages

from zcc.constants import NAME, VERSION


# The text of the README file
README = (pathlib.Path(__file__).parent / "README.md").read_text()

setup(
    name=NAME,
    version=VERSION,
    description="ZIMI ZCC helper module",
    long_description=README,
    long_description_content_type="text/markdown",
    author="Mark Hannon",
    author_email="mark.hannon@gmail.com",
    license="MIT",
    classifiers=[
        "Programming Language :: Python :: 3.12",
    ],
    packages=find_packages(exclude=["contrib", "docs", "tests"]),
    project_urls={"Source": "https://bitbucket.org/mark_hannon/zcc"},
    install_requires=[""],
)



#!/usr/bin/env python3

"""Command line parsing for zcc"""

import asyncio
import argparse
import sys


from zcc import (
    ControlPoint,
    ControlPointDescription,
    ControlPointDiscoveryService,
    ControlPointError,
)


def __options(args):
    parser = argparse.ArgumentParser(
        prog="zcc",
        description="""
                                     A command line interface to a ZIMI home controller.

                                     Operates in two basic modes: (a) discovery mode where
                                     the local network is scanned for a ZCC and the details
                                     are printed, or (b) an execute mode where an action is
                                     sent to a device""",
    )

    parser.add_argument(
        "--verbosity",
        action="store",
        type=int,
        default=0,
        help="verbosity level between 0 and 2",
    )

    command_group = parser.add_argument_group("command")

    cxg = command_group.add_mutually_exclusive_group(required=True)
    cxg.add_argument(
        "--discover", action="store_true", help="discover ZCC and print devices"
    )
    cxg.add_argument(
        "--discover-all", action="store_true", help="discover all ZCC devices on LAN"
    )
    cxg.add_argument(
        "--discover-first",
        action="store_true",
        help="discover first ZCC devices on LAN",
    )
    cxg.add_argument(
        "--execute", action="store_true", help="execute an action on a device"
    )
    cxg.add_argument(
        "--test-connection", action="store_true", help="test the connection to a device"
    )

    host_group = parser.add_argument_group("host")
    host_group.add_argument("--host", action="store", help="zcc host name|address")
    host_group.add_argument("--port", action="store", type=int, help="zcc port")
    host_group.add_argument(
        "--timeout", action="store", type=int, default=3, help="zcc timeout value"
    )

    device_group = parser.add_argument_group("device")
    device_group.add_argument("--device", action="store", help="device identifier")
    device_group.add_argument(
        "--action",
        action="store",
        type=str,
        choices=[
            "CloseDoor",
            "OpenDoor",
            "TurnOn",
            "TurnOff",
            "OpenToPercentage",
            "SetBrightness",
            "SetFanSpeed",
        ],
    )
    device_group.add_argument(
        "--value", action="store", help="device action value (for actions that require)"
    )

    return parser.parse_args(args)


async def main(args):
    """Main function."""

    options = __options(args)

    if options.discover_all:
        await ControlPointDiscoveryService(verbosity=options.verbosity).discovers()
        return

    if options.discover_first:
        await ControlPointDiscoveryService(verbosity=options.verbosity).discover()
        return

    if options.test_connection and options.host and options.port:
        result = await ControlPointDiscoveryService(
            verbosity=options.verbosity
        ).validate_connection(host=options.host, port=options.port)
        print(result)
        return

    if options.host and options.port:
        description = ControlPointDescription(host=options.host, port=options.port)
    else:
        description = await ControlPointDiscoveryService(
            verbosity=options.verbosity
        ).discover()

    controller = ControlPoint(description=description, verbosity=options.verbosity)

    await controller.connect()

    if options.discover:
        controller.print_description()

    if options.execute:
        if options.action and options.device:
            device = controller.devices.get(options.device, None)

            if device:
                if "CloseDoor" in options.action:
                    await device.close_door()
                elif "OpenDoor" in options.action:
                    await device.open_door()
                elif "TurnOn" in options.action:
                    await device.turn_on()
                elif "TurnOff" in options.action:
                    await device.turn_off()
                elif "OpenToPercentage" in options.action:
                    if options.value:
                        await device.open_to_percentage(options.value)
                    else:
                        raise ControlPointError("OpenToPercentage requires --value")
                elif "SetBrightness" in options.action:
                    if options.value:
                        await device.set_brightness(options.value)
                    else:
                        raise ControlPointError("SetBrightness requires --value")
                elif "SetFanSpeed" in options.action:
                    if options.value:
                        await device.set_fanspeed(options.value)
                    else:
                        raise ControlPointError("SetFanSpeed requires --value")
                else:
                    raise ControlPointError("No valid --device and --action")
            else:
                raise ControlPointError(f"No device {options.device} exists")
        else:
            raise ControlPointError("--execute needs --device and --action options")

    if controller:
        controller.disconnect()


if __name__ == "__main__":
    try:
        asyncio.run(main(sys.argv[1:]))
    except ControlPointError as error:
        print(f"zcc {' '.join(sys.argv[1:])}: {error}")



'''Test Basic Controller functionality.'''
import pytest

from zcc.controller import ControlPoint, ControlPointError
from zcc.device import ControlPointDevice

from tests.test_server import test_server


def test_controller_discover(test_server):
    '''Test for connection to a Controller connected to the Test Server'''

    controller = ControlPoint(timeout=1)

    assert controller.ready is True
    assert controller.host is not None
    assert controller.port == 5003
    assert controller.brand == 'zimi'
    assert len(controller.devices) == 31
    assert len(controller.doors) == 0
    assert len(controller.fans) == 0
    assert len(controller.lights) == 20
    assert len(controller.outlets) == 0


def test_controller_discover_timeout():
    '''Test if a UDP discovery times out - goes out over wire'''

    try:
        ControlPoint(timeout=1)
        assert False
    except ControlPointError as error:
        assert error.args[0] == '__init() failed - unable to discover and connect to ZCC'


def test_controller_device(test_server):
    '''Test turn_on works for valid devices'''

    controller = ControlPoint(timeout=1)

    light = controller.lights[0]

    assert light.is_off() is not True
    assert light.is_on() is True

    assert light.is_opening is not True
    assert light.is_open is not True
    assert light.location == "lounge LED strip/Lounge"
    assert light.name == "lounge LED strip"

    with pytest.raises(ControlPointDeviceError):
        light.open_door()

    with pytest.raises(ControlPointDeviceError):
        light.close_door()

    light.turn_on()

    light.turn_off()
